-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Matching.ConstraintInclusion
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- Treating NonCapture- or BV-Constraints and Subsumption on them. 
-- Required for joining critical pairs.
--
-----------------------------------------------------------------------------

module LRSX.Matching.ConstraintInclusion where
import Data.Function
import Data.Maybe
import Data.List
import qualified Data.Set as Set
import qualified Data.Map.Strict as Map
--
import Debug.Trace
--
import LRSX.Language.Syntax
import LRSX.Util.Util
import LRSX.Unification.UnificationProblem
-- =====================================================
data Domain a b = Domain a | Codomain a |  Comp b
 deriving(Eq,Ord,Show)
data Dom = Vars SyntaxComponent         -- All Variables of a Component
--          | Vars SyntaxComponent     -- Free Variables of a Component
--          | BoundVars SyntaxComponent    -- Bound variable of a Component
         | Cod SubstItem                -- Codomain of a fresh renaming 
         | SingleVar VarLift            -- A single variable
         | SingleVarCod VarLift         -- A single variable which is fresh
  deriving(Eq,Show)
  
isFreshCod (SingleVarCod _) = True
isFreshCod (Cod _) = True
isFreshCod _ = False

bvAbstractNew :: [(SyntaxComponent, SubstItem)] -> SyntaxComponent -> [Dom]  
bvAbstractNew alphabase (SCExpr (MetaCtxtSubst subst srt name e)) = error "this should not occurs, since the syntax is already splitted" $
    let ctxtvars = 
            case getFreshness (SCCtxt (MetaCtxt subst srt name)) alphabase of
                (e',Nothing,substlist) ->    (Vars e'):(map process substlist)
                (e',Just alpha,substlist) -> [(Cod alpha)] -- other substitutions can't influence the boundvars, since alpha is a fresh renaming of e'
    in ctxtvars ++ (bvAbstractNew alphabase (SCExpr e))
 where    
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  

bvAbstractNew alphabase (SCExpr (ConstantCtxtSubst subst srt name e)) = error "this should not occurs, since the syntax is already splitted" $
    let ctxtvars = 
            case getFreshness (SCCtxt (ConstantCtxt subst srt name)) alphabase of
                (e',Nothing,substlist) ->    (Vars e'):(map process substlist)
                (e',Just alpha,substlist) -> [(Cod alpha)] -- other substitutions can't influence the boundvars, since alpha is a fresh renaming of e'
    in ctxtvars ++ (bvAbstractNew alphabase (SCExpr e) )               
 where    
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  

bvAbstractNew alphabase syntaxcomp@(SCVar v) = 
  case lookup (removeSubst v) [(removeSubst t,y) | (y,SubstVar f t) <- alphabase] of  
   -- v is a fresh variable introduced by alpha renaming, since these variables cannot be renamed again, we can drop substitutions
   Just _ -> case lookup v [(t,y) | (y,SubstVar f t) <- alphabase] of 
               Nothing -> error "this is impossible?"
               _ -> [SingleVarCod v]
   Nothing ->  case getFreshness syntaxcomp alphabase of
                (e',Nothing,substlist) ->    (boundVarsOfSCComp e'):(map process substlist)
                (e',Just alpha,substlist) -> error "does not occur?" $ [(Cod alpha)] -- other substitutions can't influence the boundvars, since alpha is a fresh renaming of e'
 where    
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  
    removeSubst (MetaVarLift subst v) = MetaVarLift emptySubstitution v
    removeSubst (ConstantVarLift subst v) = ConstantVarLift emptySubstitution v
    removeSubst (VarLift subst v) = VarLift emptySubstitution v
    
    
bvAbstractNew alphabase syntaxcomp =
  case getFreshness syntaxcomp alphabase of
     (e',Nothing,substlist) ->    (boundVarsOfSCComp e'):(map process substlist)
     (e',Just alpha,substlist) -> [(Cod alpha)] -- other substitutions can't influence the boundvars, since alpha is a fresh renaming of e'
 where    
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  
    
varsAbstractNew alphabase (SCExpr (MetaCtxtSubst subst srt name e)) = error "this should not occurs, since the syntax is already splitted" $
    let ctxtvars = 
            case getFreshness (SCCtxt (MetaCtxt subst srt name)) alphabase of
                (e',Nothing,substlist) ->    (Vars e'):(map process substlist)
                (e',Just alpha,substlist) -> [Vars e', (Cod alpha)] ++ (map process substlist)
    in ctxtvars ++ (varsAbstractNew alphabase (SCExpr e))
 where    
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  
    
varsAbstractNew alphabase (SCExpr (ConstantCtxtSubst subst srt name e)) = error "this should not occurs, since the syntax is already splitted" $
    let ctxtvars = 
            case getFreshness (SCCtxt (ConstantCtxt subst srt name)) alphabase of
                (e',Nothing,substlist) ->    (Vars e'):(map process substlist)
                (e',Just alpha,substlist) -> [Vars e', (Cod alpha)] ++ (map process substlist)
    in ctxtvars ++ (varsAbstractNew alphabase (SCExpr e))
 where    
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  
  
  
  
varsAbstractNew alphabase syntaxcomp@(SCVar v) =
  case lookup (removeSubst v) [(removeSubst t,y) | (y,SubstVar f t) <- alphabase] of    
   Just _ -> case lookup v [(t,y) | (y,SubstVar f t) <- alphabase] of
               Nothing -> error "should not happen" 
               _ ->    if null substlistv then (SingleVarCod v):(map process substlistv) else error $ "should not happen: " ++ show syntaxcomp
   Nothing ->  case getFreshness syntaxcomp alphabase of
                        (e',Nothing,substlist) ->    (varsOfSCComp e'):(map process substlist)
                        (e',Just alpha,substlist) -> [freeVarsOfSCComp e', (Cod alpha)] ++ (map process substlist)
                                  -- since alpha makes e' fresh, all other substitution can only rename free variables, no bound variables
 where    
    substlistv = case v of 
                   (VarLift (Substitution s) _) -> s
                   (MetaVarLift (Substitution s) _) -> s
                   (ConstantVarLift (Substitution s) _) -> s
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  
    removeSubst (MetaVarLift subst v) = MetaVarLift emptySubstitution v
    removeSubst (ConstantVarLift subst v) = ConstantVarLift emptySubstitution v
    removeSubst (VarLift subst v) = VarLift emptySubstitution v
    
varsAbstractNew alphabase syntaxcomp =
  case getFreshness syntaxcomp alphabase of
     (e',Nothing,substlist) ->    (varsOfSCComp e'):(map process substlist)
     (e',Just alpha,substlist) -> [freeVarsOfSCComp e', (Cod alpha)] ++ (map process substlist)
                                  -- since alpha makes e' fresh, all other substitution can only rename free variables, no bound variables
 where    
    process (SubstName (None comp) alphaname) = Cod (SubstName (None comp) alphaname)
    process (SubstName (CapVars comp) alphaname) = Cod (SubstName (CapVars comp) alphaname)
    process (SubstName (LetVars comp) alphaname) = Cod (SubstName (LetVars comp) alphaname)
    process (SubstSet  renamings) = Cod (SubstSet renamings)
    process (SubstVar v1 v2) = SingleVarCod v2  
  
varsOfSCComp e@(SCVar v) = SingleVar v
varsOfSCComp e = Vars e

freeVarsOfSCComp e@(SCVar v) = SingleVar v
freeVarsOfSCComp e = Vars e

boundVarsOfSCComp e@(SCVar v) = SingleVar v
boundVarsOfSCComp e = Vars e
        
getFreshness (SCExpr (MetaExpression (Substitution sublist) e)) alphabase = 
 case
    [((SCExpr (MetaExpression (Substitution prefix) e)),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCExpr (MetaExpression (Substitution prefix) e)) ifix alphabase
    ]
  of
        [] -> (SCExpr (MetaExpression emptySubstitution e), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
        
getFreshness (SCExpr (Constant (Substitution sublist) e)) alphabase = 
 case
    [((SCExpr (Constant (Substitution prefix) e)),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCExpr (Constant (Substitution prefix) e)) ifix alphabase
    ]
  of
        [] -> (SCExpr (Constant emptySubstitution e), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
        
getFreshness (SCCtxt (MetaCtxt (Substitution sublist) srt name)) alphabase = 
 case
    [((SCCtxt (MetaCtxt (Substitution prefix) srt name)),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCCtxt (MetaCtxt (Substitution prefix) srt name)) ifix alphabase
    ]
  of
        [] -> (SCCtxt (MetaCtxt emptySubstitution srt name), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh

getFreshness (SCCtxt (ConstantCtxt (Substitution sublist) srt name)) alphabase = 
 let result =    
            case
                [((SCCtxt (ConstantCtxt (Substitution prefix) srt name)),ifix,suffix) 
                    | (prefix,ifix,suffix) <- reverse (splits2 sublist)
                    , isFreshRenaming (SCCtxt (ConstantCtxt (Substitution prefix) srt name)) ifix alphabase
                ]
            of
                    [] ->                   (SCCtxt (ConstantCtxt emptySubstitution srt name), Nothing, sublist)
                    ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
 in result
            
             
getFreshness (SCMetaEnv (Env (Substitution sublist) name)) alphabase = 
 case
    [((SCMetaEnv (Env (Substitution prefix) name)),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCMetaEnv (Env (Substitution prefix)  name)) ifix alphabase
    ]
  of
        [] -> (SCMetaEnv (Env emptySubstitution name), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
        
getFreshness (SCConstantEnv (Env (Substitution sublist) name)) alphabase = 
 case
    [((SCConstantEnv (Env (Substitution prefix) name)),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCConstantEnv (Env (Substitution prefix)  name)) ifix alphabase
    ]
  of
        [] -> (SCConstantEnv (Env emptySubstitution name), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
        
getFreshness  (SCMetaChain (Substitution sublist) typ name) alphabase = 
 case
    [(SCMetaChain (Substitution prefix) typ name,ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCMetaChain (Substitution prefix) typ name) ifix alphabase
    ]
  of
        [] -> (SCMetaChain (emptySubstitution) typ name, Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
        
getFreshness  (SCConstantChain (Substitution sublist) typ name) alphabase = 
 case
    [(SCConstantChain (Substitution prefix) typ name,ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCConstantChain (Substitution prefix) typ name) ifix alphabase
    ]
  of
        [] -> (SCConstantChain (emptySubstitution) typ name, Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
               
getFreshness (SCVar (VarLift (Substitution sublist) v)) alphabase = 
 case
    [(SCVar (VarLift  (Substitution prefix) v),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCVar (VarLift  (Substitution prefix) v)) ifix alphabase
    ]
  of
        [] -> (SCVar (VarLift  emptySubstitution v), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
    
    
getFreshness (SCVar (MetaVarLift (Substitution sublist) v)) alphabase = 
 case
    [(SCVar (MetaVarLift  (Substitution prefix) v),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCVar (MetaVarLift  (Substitution prefix) v)) ifix alphabase
    ]
  of
        [] -> (SCVar (MetaVarLift  emptySubstitution v), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
    
    
        
getFreshness (SCVar (ConstantVarLift (Substitution sublist) v)) alphabase = 
 case
    [(SCVar (ConstantVarLift  (Substitution prefix) v),ifix,suffix) 
        | (prefix,ifix,suffix) <- reverse (splits2 sublist)
        , isFreshRenaming (SCVar (ConstantVarLift  (Substitution prefix) v)) ifix alphabase
    ]
  of
        [] -> (SCVar (ConstantVarLift  emptySubstitution v), Nothing, sublist)
        ((e',ifix,suffix):_) -> (e',Just ifix,suffix)           -- ifix makes e' fresh
    
    
           

isFreshRenaming sc substitem alphabase 
    | Just items <- lookupAll sc alphabase
    , not (null [item | item <- items, item == substitem])                = True
    | otherwise                        = False


-- nccAbstractNew a p =
   -- let x1 = nccAbstractNew1 a p
       -- x2 = nccAbstractNew2 a p -- this Definition will be used
   -- in if x1 == Nothing && x2 == Nothing then Nothing
                                        -- else if isJust x1 && isJust x2 then x2
                                                                       -- else error $ "here:\n" ++ "old: " ++ show x1 ++ "\nnew: " ++ show x2
 {- DEAD CODE 
nccAbstractNew1 alphabase (s,d) = 
    let
        vardom_s = flattenDom $ varsAbstractNew alphabase s 
        bvdom_d  = flattenDom $ bvAbstractNew alphabase d 
        containsVarsOrBoundVarsDomD = not $ null ([() | Vars _ <- bvdom_d] ++ [() | Vars _ <- bvdom_d] ++ [() |(SingleVar _) <- bvdom_d])
        isCapturedByBVDOM_d x
          -- case bv(d) containts BV(item) or Var(item) or a variable (which is non fresh)
          | containsVarsOrBoundVarsDomD  = Just  $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) contains "
                                               ++ show ([x | x@(Vars _) <- bvdom_d] ++ [x | x@(Vars _) <- bvdom_d ] ++ [x | x@(SingleVar _) <- bvdom_d]) 
                                               ++ "\n"
                                               ++ "AlphaBase:" ++ show alphabase ++ "\n"
          -- case bv(d) contains only cod(alpha) and SingleVarCod:

        isCapturedByBVDOM_d (Vars item)                         =  Nothing     -- the bv are all fresh and thus cannot rename freevars 

        isCapturedByBVDOM_d y@(Vars item)                           =  Nothing  -- unklar?
--               Just $ "Alpha-Engine can't infer NCC: \n" 
--                                                 ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
--                                                ++ "Reason: " ++ "bv(d) is nonempty: " ++ show bvdom_d
--                                                ++ "and var(s) contains " ++ show y ++ "\n"
            
        isCapturedByBVDOM_d y@(Vars item)                      =  Nothing
--               Just $ "Alpha-Engine can't infer NCC: \n" 
--                                                 ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
--                                                ++ "Reason: " ++ "bv(d) is nonempty: " ++ show bvdom_d
--                                                ++ "and var(s) contains " ++ show y ++ "\n"
        isCapturedByBVDOM_d y@(Cod (SubstName (LetVars expr) name)) =
            if    (Cod (SubstName (LetVars expr) name)) `elem` bvdom_d
               || (Cod (SubstName (CapVars expr) name)) `elem` bvdom_d
               || (Cod (SubstName (None expr) name)) `elem` bvdom_d
            then
            Just  $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) contains "
                                               ++ show ([x | x@(Cod(SubstName (LetVars expr') name')) <- bvdom_d, expr' == expr, name' == name]
                                                       ++
                                                        [x | x@(Cod(SubstName (None expr') name')) <- bvdom_d, expr' == expr, name' == name]
                                                       ++
                                                        [x | x@(Cod(SubstName (LetVars expr') name')) <- bvdom_d, expr' == expr, name' == name])
                                               ++ "and var(s) contains " ++ show y ++ "\n"
                                               ++ "AlphaBase:" ++ show alphabase ++ "\n"
            else Nothing
        isCapturedByBVDOM_d y@(Cod (SubstName (CapVars expr) name)) =
            if    (Cod (SubstName (LetVars expr) name)) `elem` bvdom_d
               || (Cod (SubstName (CapVars expr) name)) `elem` bvdom_d
               || (Cod (SubstName (None expr) name)) `elem` bvdom_d
            then
            Just  $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) contains "
                                               ++ show ([x | x@(Cod(SubstName (LetVars expr') name')) <- bvdom_d, expr' == expr, name' == name]
                                                       ++
                                                        [x | x@(Cod(SubstName (None expr') name')) <- bvdom_d, expr' == expr, name' == name]
                                                       ++
                                                        [x | x@(Cod(SubstName (LetVars expr') name')) <- bvdom_d, expr' == expr, name' == name])
                                               ++ "and var(s) contains " ++ show y ++ "\n"
            else Nothing
        isCapturedByBVDOM_d y@(Cod (SubstName (None expr) name)) =
            if    (Cod (SubstName (LetVars expr) name)) `elem` bvdom_d
               || (Cod (SubstName (CapVars expr) name)) `elem` bvdom_d
               || (Cod (SubstName (None expr) name)) `elem` bvdom_d
            then
            Just  $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) contains "
                                               ++ show ([x | x@(Cod(SubstName (LetVars expr') name')) <- bvdom_d, expr' == expr, name' == name]
                                                       ++
                                                        [x | x@(Cod(SubstName (None expr') name')) <- bvdom_d, expr' == expr, name' == name]
                                                       ++
                                                        [x | x@(Cod(SubstName (LetVars expr') name')) <- bvdom_d, expr' == expr, name' == name])
                                               ++ "and var(s) contains " ++ show y ++ "\n"
            else Nothing
        isCapturedByBVDOM_d y@(SingleVarCod v) =
            if    (SingleVarCod v) `elem` bvdom_d
               || (SingleVar v) `elem` bvdom_d
             then
              Just $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) contains "
                                               ++ show ([x | x@(SingleVarCod n)  <- bvdom_d, n == v]
                                                       ++
                                                        [x | x@(SingleVar n) <- bvdom_d, n== v])
                                               ++ "and var(s) contains " ++ show y ++ "\n"
             else Nothing
        
        isCapturedByBVDOM_d y@(SingleVar v) 
            |   (SingleVar v) `elem` bvdom_d =
              Just $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) is nonempty: " ++ show bvdom_d
                                               ++ "and var(s) contains " ++ show y ++ "\n"
            |   (SingleVarCod v) `elem` bvdom_d =
              Just $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) is nonempty: " ++ show bvdom_d
                                               ++ "and var(s) contains " ++ show y ++ "\n"
            
            |   all isFreshCod bvdom_d         = Nothing
                
            |   otherwise                      = error "this case should not happen" $
        
              Just $ "Alpha-Engine can't infer NCC: \n" 
                                                ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
                                               ++ "Reason: " ++ "bv(d) is nonempty: " ++ show bvdom_d
                                               ++ "and var(s) contains " ++ show y ++ "\n"
            
            
--             if    (SingleVarCod v) `elem` bvdom_d
--                || (SingleVar v) `elem` bvdom_d
--              then
--               Just $ "Alpha-Engine can't infer NCC: \n" 
--                                                 ++ "NCC (s,d) = " ++ show (s,d) ++ "\n" 
--                                                ++ "Reason: " ++ "bv(d) contains "
--                                                ++ show ([x | x@(SingleVarCod n)  <- bvdom_d, n == v]
--                                                        ++
--                                                         [x | x@(SingleVar n) <- bvdom_d, n== v])
--                                                ++ "and var(s) contains " ++ show y ++ "\n"
--              else Nothing
             
        isCapturedByBVDOM_d y = error $ show y
 
    in if all (isNothing) [isCapturedByBVDOM_d item | item <- vardom_s] then
         Nothing
        else
         Just (intercalate ";" $ map fromJust $ filter isJust [isCapturedByBVDOM_d item | item <- vardom_s])
-}
flattenDom [] = []       
flattenDom (x@(Cod item):rest) =  (map getCod $ flatten [item]) ++ (flattenDom rest)
    where 
     getCod (SubstName e n)  = Cod $ SubstName e n
     getCod (SubstVar v1 v2) = SingleVarCod v2

flattenDom (x:rest) = x:(flattenDom rest)

flatten ((SubstSet set):rest) = flatten (set ++ rest)
flatten (x:rest) = x:(flatten rest)
flatten [] = []


nccAbstractNew alphabase (s,d) = 
    let
        vardom_s = flattenDom $ varsAbstractNew alphabase s 
        bvdom_d  = flattenDom $ bvAbstractNew alphabase d 
        checkset = [(i,j) | i <- vardom_s, j <- bvdom_d]
        -- x vs y  ok if x y are concrete and different
        check q@(SingleVar (VarLift _ v1)) p@(SingleVar (VarLift _ v2)) 
          | v1 == v2 = Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
          | otherwise = Nothing
        check q@(SingleVar _) p@(SingleVar _) = Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- x vs fr(x) ok
        check q@(SingleVar item) p@(SingleVarCod v) = Nothing
        -- x vs BV(I') => FAIL
        check q@(SingleVar item) p@(Vars item') = Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- x vs Cod alpha -> Ok, sinc Cod is always fresh                                                
        check q@(SingleVar item) p@(Cod _)          = Nothing
        
        -- fr(x) vs x ok
        check q@(SingleVarCod _) p@(SingleVar _) = Nothing
        
        -- fr(x) vs fr(y)  ok unless x = y
        check q@(SingleVarCod u) p@(SingleVarCod v) 
          | u == v = Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
          
          | otherwise = Nothing
        -- fr(x) vs BV(I') ok
        check q@(SingleVarCod item) p@(Vars item') = Nothing
        -- fr(x) vs Cod alpha -> Ok, sinc Cod is always fresh                                                
        check q@(SingleVarCod item) p@(Cod _)          = Nothing
        
        -- FV(I) vs x  => FAIL since x may be in FV(I)
        check q@(Vars item) p@(SingleVar v) = Just $
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- FV(I) vs fresh(x) => NOFAIL, since x is fresh w.r.t. FV(I) 
        check q@(Vars item) p@(SingleVarCod v) = Nothing
        -- FV(I) vs BV(I') => FAIL
        check q@(Vars item) p@(Vars item') = Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- FV(I) vs Cod alpha -> Ok, sinc Cod is always fresh                                                
        check q@(Vars item) p@(Cod _)          = Nothing
--
        -- BV(I) vs x  => FAIL since x may be in FV(I)
        check q@(Vars item) p@(SingleVar v) = Just $
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- FV(I) vs fresh(x) => NOFAIL, since x is fresh w.r.t. BV(I) 
        check q@(Vars item) p@(SingleVarCod v) = Nothing
        -- FV(I) vs BV(I') => FAIL
        check q@(Vars item) p@(Vars item') = Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- FV(I) vs Cod alpha -> Ok, sinc Cod is always fresh                                                
        check q@(Vars item) p@(Cod _)          = Nothing
--
        -- Vars(I) vs x  => FAIL since x may be in FV(I)
        check q@(Vars item) p@(SingleVar v) = Just $
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- Vars(I) vs fresh(x) => NOFAIL, since x is fresh w.r.t. BV(I) 
        check q@(Vars item) p@(SingleVarCod v) = Nothing
        -- Vars(I) vs BV(I') => FAIL
        check q@(Vars item) p@(Vars item') = Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
        -- Vars(I) vs Cod alpha -> Ok, sinc Cod is always fresh                                                
        check q@(Vars item) p@(Cod _)          = Nothing
--
        -- Cod(alpha) vs x => ok, since Cod is fresh and x is not fresh
        check q@(Cod item) p@(SingleVar v) = Nothing 
--         - Cod(I) vs fresh(x) => NOFAIL, since x is fresh wrt all Cods
        check q@(Cod item) p@(SingleVarCod v) = Nothing
        -- Cod(alpha) vs BV(I) ok
        check q@(Cod item) p@(Vars item') =Nothing
                                              
        -- Cod(alpha1) vs Cod(alpha2) 
        -- Fail if alpha_1 and alpha_2 are same, else ok
        
        check q@(Cod (SubstName _ name)) p@(Cod (SubstName _ name1))
          | name == name1 =     Just $ 
                                                    "Alpha-Engine can't infer NCC:\n"
                                                ++  "NCC (s,d) = " ++ show (s,d) ++ "\n"
                                                ++  "Reason: var(s) contains " ++ show q ++ " and bv(d) contains " ++ show p
          | otherwise = Nothing
                                                
         
        
 
    in if all (isNothing) [check i j | (i,j) <- checkset]
          then
         Nothing
        else
         Just (intercalate ";" $ map fromJust $ filter isJust [check i j | (i,j) <- checkset])

-- =====================================================       
-- the main interface functions are 
-- bvImplies  and nccImplies`
-- where 
--   * bvImplies checks whether for constraints (s,s') 
--     BV(s') cap Var(s)  = emptyset 
--   * nccImplies checks whether for constraints (s,d)
--     CV(d) cap Var(s) = emptyset 

bvImplies capconConst  capconVar alphabase= 
 let cconst = splitBVCon capconConst
     cvar   = splitBVCon capconVar
     process []  = ("",Nothing)
     process ((s,d):rest) =
      case Map.lookup s cconst of 
       Just d' -> if d `Set.isSubsetOf` d' then 
                    (process rest) else 
                    ("",Just ("BVCon implication" ++ show (s,d)))
       Nothing -> if isFreshSym s then process rest else
                        ("",Just ("BVCon implication" ++ show (s,d)))
 in process (Map.assocs cvar)

-- the helper function splitBVCon takes Contraints (s,s') and
-- splits them into (Var-construct,set of BV-constructs)
-- e.g. (letrec E in \X.S, letrec E1; E2 in (app Y S3))
-- results in (E,{E1,E2,S3}), (X,{E1,E2,S3}), (S,{E1,E2,S3))
-- (assume that all constructs are constants)
splitBVCon :: NonCaptureConstraints -> SplittedNCCs -- Map.Map SyntaxComponent (Set.Set SyntaxComponent)   
splitBVCon capcon =
 Map.fromList
 [(s',Set.unions $ map snd grp)
  | grp@((s',_):_) <- 
    (
     groupBy ((==) `on` fst) $
      sortBy (compare `on` fst) $
       [(s',getSCBoundVars d) | (s,d) <- capcon
                              , s' <- Set.elems (getSCAllVars s)]
    )           
 ]    

removeFreshSymbols :: SplittedNCCs -> SplittedNCCs 
removeFreshSymbols mp =
    foldr (\key oldmap -> 
                  Map.insert key (Set.filter (\x -> x == key || (not (isFreshSym x))) (mp Map.! key)) oldmap) mp (Map.keys mp)
 
nccImplies :: [EnvVar]
                            -> NonCaptureConstraints
                            -> NonCaptureConstraints
                            -> [(SyntaxComponent, SubstItem)]
                            -> (String, Maybe String)
nccImplies delta2 nablaCon deltaCon  alphabase  =
 let 
     splitNabla = (unfoldSplittedNCCs (splitCapCon nablaCon))
     splitDeltaSet = (splitCapCon deltaCon)
     splitDelta = unfoldSplittedNCCs splitDeltaSet
     
     check (s@(SCVar (VarLift (Substitution []) v)),d@(SCVar (VarLift (Substitution []) v')))
       -- case (s,d) = (x,y) where x and y are concrete variables which are different and where no substitution is involved:
       | v /= v' = ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                    ++ "\n (s,d) holds since both are different concrete variables without a substitution"
                   ,Nothing)
     -- *******************************************************************++
     -- Constraint (s,d) holds since it can be inferred from nabla 
     --  or s or d is a fresh symbol, where for the case s = d we have to check 
     --  if s and d must be instantiated with non-capturing instances      
     check (s,d) 
       | (s,d)  `elem` splitNabla = ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                                     ++ "\n (s,d) holds since (s,d) is contained in the given nabla base"
                                    ,Nothing)
       | isFreshSym s 
       , d == s
       , Nothing <- checkSC delta2 s d splitDeltaSet = ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                                   ++
                                                   "\nNCCIMPLLOG: (s,d) holds, since s is a fresh symbol"
                                                   ,Nothing)
       | isFreshSym s 
       , d /= s                                 = ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                                   ++
                                                   "\nNCCIMPLLOG: (s,d) holds, since s is a fresh symbol"
                                                   ,Nothing)
       | isFreshSym d 
       , d /= s                                 = ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                                   ++
                                                   "\nNCCIMPLLOG: (s,d) holds, since d is a fresh symbol"
                                                   ,Nothing)       
     -- *************************************************************              
     -- SEVERAL CHECKS WHETHER THE CONSTRAINT (s,d) HOLDS ALREADY,
     -- SINCE ALL GROUND INSTANTIATIONS WHICH VIOLATE (s,d) DO NOT
     -- FULFILL THE WEAK DVC 
     -- Only cases where red (i.e constant) variables occur,
     -- since blue (i.e. meta) variables are already handled above
     -- wrong!!! since other constant variables may affect the weak DVC\n
     -- eg (X,x)    (  app C[var x]  \X.X   =?= app C[var x] \X.X
     --    X |-> x violates the DVC, but if x in \CV(C) then the DVC is fulfilled     
     --     
     -- (X,x)  holds if (X |-> x)  violates the DVC  
     --
     -- check (s@(SCVar var@(ConstantVarLift (Substitution []) v)),d@(SCVar var'@(VarLift (Substitution []) v')))
        -- | Just err <- checkRhoPrime [MapVarLift (var :|-> var')] = 
             -- ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                                     -- ++ "\n (s,d) holds since  mapping s |-> d violates the weak DVC\n" ++ err
                                    -- ,Nothing)
     --     
     -- (x,X)  holds if (X |-> x)  violates the DVC  
     --
     -- check (s@(SCVar var'@(VarLift (Substitution []) v)),d@(SCVar var@(ConstantVarLift (Substitution []) v')))
        -- | Just err <- checkRhoPrime [MapVarLift (var :|-> var')] = 
             -- ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                                     -- ++ "\n (s,d) holds since  mapping d |-> s violates the weak DVC\n" ++ err
                                    -- ,Nothing)
     --                               
     -- (X,Y) holds if (X |-> y), (Y |-> y) where y is fresh violates the DVC 
     --  since for all other instantiations the constraint clearly holds
     --     
     -- check (s@(SCVar var@(ConstantVarLift (Substitution []) v)),d@(SCVar var'@(ConstantVarLift (Substitution []) v')))
        -- | Just err <- let freshvar = (VarLift emptySubstitution ("x_" ++ f)) 
                      -- in checkRhoPrime [MapVarLift (var :|-> freshvar),MapVarLift (var' :|-> freshvar)] = 
             -- ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                                     -- ++ "\n (s,d) holds since  mapping s |-> y, d |-> y violates the weak DVC\n" ++ err
                                    -- ,Nothing)
     --                               
     -- (x,D) holds if all instantiations D -> d where x in CV(d) violate the DVC 
     --  since for all other instantiations the constraint clearly holds.
     --  It suffices to test D -> letrec x=[.] in c  since if this test fails, then also other instantiation with 
     --  x in CV(d) will fail the weak DVC test    
     -- check (s@(SCVar var@(VarLift (Substitution []) v)),d@(SCCtxt d'@(ConstantCtxt sub srt name)))
        -- |  Just err <-  checkRhoPrime [MapExpr (d' :|-> Letrec empty{bindings=[var:=Hole emptySubstitution]} (Fn "__dummyconstant" []))] = 
             -- ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                                     -- ++ "\n (s,d) holds since  mapping d |-> letrec s=[.] in c violates the weak DVC\n" ++ err
                                    -- ,Nothing)
                                    
     -- ****************************************************************
                                    
     -- (E,v) in Delta & (E,E) in nabla
     check (e@(SCConstantEnv (Env (Substitution []) _)),d)
       | (e,e) `elem` splitNabla  = ("NCCIMPLLOG:  processing (s,d) = " ++ show (e,d)
                                     ++ "\n (s,d) holds since (d,d) is contained in the given nabla base"
                                    ,Nothing)
     -- (CC,v) in Delta & (CC,CC) in nabla and sort(CC) = TRIV
     check (cc@(SCConstantChain (Substitution []) VarChain name),d)
       | (cc,cc) `elem` splitNabla  = ("NCCIMPLLOG:  processing (s,d) = " ++ show (cc,d)
                                     ++ "\n (s,d) holds since (d,d) is contained in the given nabla base"
                                    ,Nothing)
     -- (u,E) in Delta & (E,E) in nabla
     check (s,e@(SCConstantEnv (Env (Substitution []) _)))
       | (e,e) `elem` splitNabla  = ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,e)
                                     ++ "\n (s,d) holds since (d,d) is contained in the given nabla base"
                                    ,Nothing)
     -- (u,CC) in Delta & (CC,CC) in nabla
     check (s,cc@(SCConstantChain (Substitution []) typ name))
       | (cc,cc) `elem` splitNabla  = ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,cc)
                                     ++ "\n (s,d) holds since (d,d) is contained in the given nabla base"
                                    ,Nothing)
     -- (u,D) in Delta & (D,D) in nabla ? ja da CV(D) = emptyset 
     check (s,d@(SCCtxt (ConstantCtxt (Substitution []) srt name)))
       | (d,d) `elem` splitNabla
       , srt > ACtxt = ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                       ++ "\n (s,d) holds since (d,d) is contained in the given nabla base"
                       ,Nothing)
     -- (x,U) and (U,x) in nabla for U = x,X,D,E,Ch,
     check (s,d)
      | case s of 
         (SCVar (ConstantVarLift (Substitution []) v)) -> True
         (SCVar (VarLift (Substitution []) v)) -> True
         _ -> False
      , case d of 
         (SCVar (ConstantVarLift (Substitution []) v)) -> True
         (SCVar (VarLift (Substitution []) v)) -> True
         (SCCtxt (ConstantCtxt (Substitution []) _ _)) -> True
         (SCConstantEnv (Env (Substitution []) _)) -> True
         (SCConstantChain (Substitution []) _ _) -> True
         _ -> False
      , s /= d
      , (d,s) `elem` splitNabla  = ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                                    ++ "\n (s,d) holds since (d,s) is contained in the given nabla base"
                                    ,Nothing)
     -- (Ch,U) and (U,Ch) in nabla for U = x,X,D,E,Ch',  where class(Ch)  = Triv
     check (s,d)
      | case s of 
         (SCConstantChain (Substitution []) VarChain _) -> True
         _ -> False
      , case d of 
         (SCVar (ConstantVarLift (Substitution []) v)) -> True
         (SCVar (VarLift (Substitution []) v)) -> True
         (SCCtxt (ConstantCtxt (Substitution []) _ _)) -> True
         (SCConstantEnv (Env (Substitution []) _)) -> True
         (SCConstantChain (Substitution []) _ _) -> True
         _ -> False
      , s /= d
      , (d,s) `elem` splitNabla  = ("NCCIMPLLOG:  processing (s,d) = " ++ show (s,d)
                                    ++ "\n (s,d) holds since (d,s) is contained in the given nabla base"
                                    ,Nothing)
                       
                       
                       

     check (s,d)  
       | isNothing (nccAbstractNew alphabase (s,d)) =            
                                    ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                    ++                          
                                    "\nNCCIMPLLOG: capture constraint (s,d) is not implied by constraint-base but by alpha base " ++ show alphabase 
                                    ,Nothing)
       | otherwise =        let str =      "\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                        ++ "\nNCCIMPLLOG: Capture Constraint not implied by constraint-base: "
                                        ++"\nNCCIMPLLOG: Constraint Base:" ++  show (splitNabla)
                                        ++"\nNCCIMPLLOG: Needed constraints: " ++ show (splitDelta)
                                        ++"\nNCCIMPLLOG: Alpha-Engine:\n"
                                        ++ show (nccAbstractNew alphabase (s,d)) 
                                        ++ "\nNCCIMPLLOG: STOPPED WITHOUT SUCCESS" 
                            in (str, Just str)
       
     allchecks = map check splitDelta
 in 
  if all isNothing (map snd allchecks) then 
      let (s,t) = (intercalate "\n" $ map fst allchecks, Nothing) in 
         -- trace s 
         (s,t)
   else (head $ filter (isJust . snd) allchecks)
   
-- nccImplies delta2 nablaC deltaC alphabase =
 -- case  nccImpliesOld delta2 nablaC deltaC alphabase  of
  -- (str,Nothing) -> case nccImpliesNew delta2 nablaC deltaC alphabase of
                    -- (str',Nothing) -> (str',Nothing)
                    -- (str',Just err) -> error $ "new check delivers Just / old delivers Nothing" ++ "\nOld: " ++ show str ++ "\n\nNEW:" ++ str'
  -- (str,Just _) -> case nccImpliesNew delta2 nablaC deltaC alphabase of
                    -- (str',Nothing)  -> (error $ "new check delivers Nothing / old delivers Just" ++ "\nOld: " ++ show str ++ "\n\nNEW:" ++ str')
                    -- (str',Just err) -> (str',Just err)
                    

{- DEAD CODE
nccImpliesOld delta2 capconConst capconVar alphabase =
 let cconst = splitCapCon capconConst
     cvar   = removeFreshSymbols $
               splitCapCon capconVar
     process []  = ("\nNCCIMPLLOG: STOPPED SUCCESSFULLY\n", Nothing)
     --
     -- old rules:
     --
--      process ((s@(SCVar (VarLift (Substitution []) v)),d):rest)  
--           | all (isConcreteSCVar) (Set.elems d)
--           , not (Set.member s d) = 
--               let (str,r) =  process rest
--               in ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
--                   ++
--                   "\nNCCIMPLLOG: (s,d) holds since all elements in d are concrete variables and s is not an element of d"
--                   ++ str
--                  ,r)
-- --                   
--              
--               
{-   
     process ((s@((SCVar (ConstantVarLift sub v))),d):rest) 
          | all (isConcreteSCVar) (Set.elems d)
          , not (null (Set.elems d))
          , not (Set.member (SCVar $ ConstantVarLift (Substitution []) v) d) = 
              let (str,r) =  process rest
                  stri = 
                    "\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                    ++
                    "\nNCCIMPLLOG: (s,d) holds since all elements in d are concrete variables and " ++ show (SCVar $ ConstantVarLift (Substitution []) v) ++ "  is not an element of d"
                  -- ++ str
              in trace stri (stri++str,r)
    ---
    ---
    ---
  -}  
                
     process ((s@(SCVar (VarLift (Substitution []) v)),d):rest)
       -- case (s,d) = (x,y) where x and y are concrete variables which are different and where no substitution is involved:
       | all (isConcreteSCVar) (Set.elems d)
       , null [() | (SCVar (VarLift (Substitution sub) v)) <- Set.elems d, not (null sub)]
       , not (Set.null d)
       , not (Set.member s d) = process rest
     process ((s,d):rest) 
         -- does this ever happen?
         | null (Set.elems d) = 
             let (str,r) =  process rest
             in ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                 ++
                 "\nNCCIMPLLOG: (s,d) holds since d is empty"
                 ++ str
                ,r)
     process ((s,d):rest) =
      case Map.lookup s cconst of 
       Just d' -> 
            -- remove all (\eta s,\eta d) which are also in nabla
            if d `Set.isSubsetOf` d' 
               then
                    let (str,r) =  process rest
                    in ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                            ++
                            "\nNCCIMPLLOG: (s,d) holds since (s,d') in constraint base"
                            ++
                            "\nNCCIMPLLOG: where d'=" ++ show d'
                            ++
                            "\nNCCIMPLLOG: and d is a subset of d"
                            ++ str
                        ,r)
               else let 
                        
                        remaining = (Set.elems (d Set.\\ d'))
                    
                        
                    in
                      if all isNothing (map (\x -> nccAbstractNew alphabase (s,x)) remaining)
                        then if null (Set.elems (d Set.\\ d')) 
                                then error "STOPPED: in process(l.701 of ConstraintInclusion" 
                                else let (str,r) = process rest
                                     in   ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                            ++                          
                                            "\nNCCIMPLLOG: capture constraint (s,d) is not implied by constraint-base but by alpha base "
                                            ++
                                            "\nNCCIMPLLOG: " ++ show [(s,c) | c <- remaining]
                                            ++
                                            "\nNCCIMPLLOG:Constraint Base:" ++  show cconst
                                            ++
                                            "\nNCCIMPLLOG:Needed constraints: " ++ show cvar
                                            ++
                                            "\nNCCIMPLLOG: Alpha-Engine runs:\n"
                                            ++ (unlines  ["NCCIMPLLOG:\t(s,c):" ++ show (s,d) ++ 
                                                "NCCIMPLLOG: \tvar(s)=" ++ show(flattenDom $ varsAbstractNew alphabase s) ++ 
                                                "\nNCCIMPLLOG:\tbv(c) =" ++ show(flattenDom $ bvAbstractNew alphabase c) | c <- remaining])
                                                ++ (intercalate ";" ( map fromJust (filter isJust (map (\x -> nccAbstractNew alphabase (s,x)) remaining))))
                                            ++ str 
                                            ,r)
                        else    
                            let str =      "\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                        ++ "\nNCCIMPLLOG: Capture Constraints not implied by constraint-base: "
                                        ++ "\nNCCIMPLLOG: " ++ show [(s,c) | c <- (Set.elems (d Set.\\ d'))]
                                        ++"\nNCCIMPLLOG: Constraint Base:" ++  show cconst
                                        ++"\nNCCIMPLLOG: Needed constraints: " ++ show cvar
                                        ++"\nNCCIMPLLOG: Alpha-Engine:\n"
                                        ++ (intercalate ";" ( map fromJust (filter isJust (map (\x -> nccAbstractNew alphabase (s,x)) (Set.elems (d Set.\\ d'))))))
                                        ++ "\nNCCIMPLLOG: STOPPED WITHOUT SUCCESS" 
                            in (str, Just str)
       Nothing -> 
                   if isFreshSym s 
                      then 
                        if  (s `Set.member` d) 
                          then
                              case  checkSC delta2 s d capconVar of
                                Nothing -> 
                                  let (str,r) = process rest
                                  in ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                      ++
                                      "\nNCCIMPLLOG: (s,d) holds, since s is a fresh symbol"
                                      ++ str,r)
                                Just err -> ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d) ++ "NCCIMPLLOG: fails:" ++  err,  Just err)
                          else
                              let (str,r) = process rest
                              in ("\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                  ++
                                  "\nNCCIMPLLOG: (s,d) holds, since s is a fresh symbol"
                                  ++ str,r)
                      else                    
                        if all isNothing (map (\x -> nccAbstractNew alphabase (s,x)) (Set.elems (d))) 
                            then if null (Set.elems d) 
                                    then error "stopped2" 
                                    else
                                        let (str,r) = process rest
                                        in (
                                            "\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                            ++
                                            "\nNCCIMPLLOG:Capture Constraints not implied by constraint-base but by alpha base: "
                                            ++
                                            "\nNCCIMPLLOG:" ++ show [(s,c) | c <- (Set.elems (d))]
                                            ++"\nNCCIMPLLOG:Constraint Base:" ++  show cconst
                                            ++"\nNCCIMPLLOG:Needed constraints: " ++ show cvar
                                            ++"\nNCCIMPLLOG:Alpha-Engine:\n"
                                            ++ (unlines ["\nNCCIMPLLOG: \t(s,c):" ++ show (s,c) ++ 
                                                "\nNCCIMPLLOG:\tvar(s)=" ++ show(flattenDom $ varsAbstractNew alphabase s) ++ 
                                                "\nNCCIMPLLOG:\tbv(c) =" ++ show(flattenDom $ bvAbstractNew alphabase c) | c <- (Set.elems (d))])
                                            ++  (intercalate ";" (map fromJust $ filter isJust (map (\x -> nccAbstractNew alphabase (s,x)) (Set.elems (d)))))
                                            ++ str, r)
                            else
                                    let str =
                                            "\nNCCIMPLLOG: processing (s,d) = " ++ show (s,d)
                                            ++ 
                                            "\nNCCIMPLLOG: Capture Constraints not implied by constraint-base: "
                                            ++ 
                                            "\nNCCIMPLLOG: " ++ show [(s,c) | c <- (Set.elems (d))]
                                            ++
                                            "\nNCCIMPLLOG: Constraint Base:" ++  show cconst
                                            ++"\nNCCIMPLLOG: Needed constraints: " ++ show cvar
                                            ++"\nNCCIMPLLOG:Alpha-Engine:\n"
                                            ++  (intercalate ";" (map fromJust $ filter isJust (map (\x -> nccAbstractNew alphabase (s,x)) (Set.elems (d)))))
                                            ++ "\nNCCIMPLLOG: STOPPED WITHOUT SUCCESS" 
                                    in  -- trace str 
                                      (str, Just str)
 in 
     process (Map.assocs cvar)
-}     
     
-- similar to splitBVCon, splitCapCon takes Contraints (s,d) and
-- splits them into (Var-construct,set of Capturing-constructs)
-- e.g. (letrec E in \X.S, letrec E1; E2 in (app Y [.]))
-- results in (E,{E1,E2}), (X,{E1,E2}), (S,{E1,E2))
-- (assume that all constructs are constants)

type SplittedNCCs =  Map.Map SyntaxComponent (Set.Set SyntaxComponent)   

checkNCCs delta2 capcon =
 let m = Map.assocs $ splitCapCon capcon 
     go ((s,d):rest)
      | (Set.member s d) = case checkSC delta2 s d capcon of 
                              Just str -> Just str
                              Nothing -> go rest
      | otherwise = go rest
     go [] = Nothing
  in go m

checkSC delta2 s d capcon = 
          case s of 
           (SCVar v) ->  Just ("The NCC: (s,d) = " ++ show (s,d) ++ "fails, since s occurs in d and s  \n Original NCCs:" ++ show capcon)
           (SCExpr (VarL v)) -> error "should not occur"
           (SCExpr (Fn f _)) -> error "should not occur"
           (SCExpr (Letrec _ _)) -> error "should not occur"
           (SCExpr (MetaExpression _ _)) -> Nothing
           (SCExpr (Constant _ _)) -> Nothing     
           (SCExpr (Hole _)) -> error "should not occur"
           (SCExpr (MetaCtxt sub srt name)) -> error "should not occur"
           (SCExpr (MetaCtxtSubst sub srt name _)) -> error "should not occur"
           (SCExpr (ConstantCtxt sub srt name)) -> error "should not occur"
           (SCExpr (ConstantCtxtSubst _ _ _  _)) -> error  "should not occur"
           (SCCtxt (VarL v)) -> error "should not occur"
           (SCCtxt (Fn f _)) -> error "should not occur"
           (SCCtxt (Letrec _ _)) -> error "should not occur"
           (SCCtxt (MetaExpression _ _)) -> error "should not occur"
           (SCCtxt (Constant _ _)) -> error "should not occur"
           (SCCtxt (Hole _)) -> error "should not occur"
           (SCCtxt (MetaCtxt sub srt name)) -> Nothing
           (SCCtxt (MetaCtxtSubst sub srt name _)) -> error "should not occur"
           (SCCtxt (ConstantCtxt sub srt name)) -> Nothing
           (SCCtxt (ConstantCtxtSubst _ _ _ _)) -> error  "should not occur"
           (SCMetaEnv (Env sub name)) ->  
             if (Env sub name) `elem` delta2 || (Env emptySubstitution name) `elem` delta2 then
               Just ("The NCC: (s,d) = " ++ show (s,d) ++ "fails, since s occurs in d and s in non-empty \n Original NCCs:" ++ show capcon)
                                                                                           else 
               Nothing
           (SCConstantEnv (Env sub name)) ->  
             if (Env sub name) `elem` delta2 || (Env emptySubstitution name) `elem` delta2 then
               Just ("The NCC: (s,d) = " ++ show (s,d) ++ "fails, since s occurs in d and s in non-empty \n Original NCCs:" ++ show capcon)
                                                                                           else 
               Nothing
           (SCMetaChain sub typ name) -> Nothing
           (SCConstantChain sub typ name) -> Nothing
  
  
splitCapCon :: NonCaptureConstraints -> SplittedNCCs
splitCapCon capcon =
 Map.fromList
 [(s',Set.unions $ map snd grp)
  | grp@((s',_):_) <- 
    (
     groupBy ((==) `on` fst) $
      sortBy (compare `on` fst) $
       [(s',getSCCapVars d) | (s,d) <- capcon
                              , s' <- Set.elems (getSCAllVars s)]
    )           
 ]    
 
unfoldSplittedNCCs mp = 
 nub [(s,u) | (s,dset) <- Map.assocs mp, u <- Set.elems dset] 
 

-- a show function for splitted constraints:
showSplitted mp = unlines $ map (\x -> "    " ++ x) $ lines $ "\n{\n" ++ (unlines $ map show [(d, Set.elems set) | (d,set) <- Map.assocs mp]) ++ "}\n"
  

-- conversion from splitted NCCs into usual NCCs by forming expressions and contexts
splitToNCC :: SplittedNCCs -> NonCaptureConstraints
splitToNCC mp =
   [(Left (scItemToExpr t),scItemToScope d)|(t,dset) <- Map.assocs mp, d <- Set.elems dset]
 
splitFlatToNCC list = 
   [(Left (scItemToExpr t),scItemToScope d)|(t,d) <- list]
   
scItemToScope (SCVar v) = Letrec 
                            empty{bindings = [v := (Fn "__internalc" [])]} 
                            (Hole (Substitution [])) 
scItemToScope (SCExpr e) = e 
scItemToScope (SCCtxt c) = c 
scItemToScope (SCMetaEnv e) = Letrec empty{metaEnvVars=[e]} (Hole (Substitution []))
scItemToScope (SCConstantEnv e) = Letrec empty{constantEnvVars=[e]} (Hole (Substitution []))
scItemToScope (SCMetaChain sub typ name) = 
    Letrec 
        empty{metaChains=[MetaChain 
                            sub 
                            typ 
                            name 
                            (VarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
        (Fn "__internalc" [])
scItemToScope (SCConstantChain sub typ name) = 
    Letrec 
        empty{constantChains=[ConstantChain 
                            sub 
                            typ 
                            name 
                            (VarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
        (Fn "__internalc" [])
                            
  
scItemToExpr (SCVar v)         = VarL v
scItemToExpr (SCExpr e)        = e 
scItemToExpr (SCCtxt c)        = c
scItemToExpr (SCMetaEnv e)     = Letrec empty{metaEnvVars=[e]} (Fn "__internalc" [])
scItemToExpr (SCConstantEnv e) = Letrec empty{constantEnvVars=[e]} (Fn "__internalc" [])
scItemToExpr (SCMetaChain sub typ name) = 
    Letrec 
        empty{metaChains=[MetaChain 
                            sub 
                            typ 
                            name 
                            (VarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
        (Fn "__internalc" [])
scItemToExpr (SCConstantChain sub typ name) = 
    Letrec 
        empty{constantChains=[ConstantChain 
                            sub 
                            typ 
                            name 
                            (VarLift (Substitution []) ("__internalvar_" ++ name))
                            (Fn "__internalc" [])]}
        (Fn "__internalc" [])

  

   
   
